# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Advanced Technology Microwave Sounder (ATMS) Level 1B product reader.

The format is explained in the `ATMS L1B Product User Guide`_

.. _`ATMS L1B Product User Guide`:
  https://docserver.gesdisc.eosdis.nasa.gov/public/project/Sounder/ATMS_V3_L1B_Product_User_Guide.pdf

"""

import datetime as dt
import logging

from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)

DATE_FMT = "%Y-%m-%dT%H:%M:%SZ"


class AtmsL1bNCFileHandler(NetCDF4FileHandler):
    """Reader class for ATMS L1B products in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize file handler."""
        super().__init__(
            filename, filename_info, filetype_info, auto_maskandscale=True,
        )

    @property
    def start_time(self):
        """Get observation start time."""
        return dt.datetime.strptime(self["/attr/time_coverage_start"], DATE_FMT)

    @property
    def end_time(self):
        """Get observation end time."""
        return dt.datetime.strptime(self["/attr/time_coverage_end"], DATE_FMT)

    @property
    def platform_name(self):
        """Get platform name."""
        return self["/attr/platform"]

    @property
    def sensor(self):
        """Get sensor."""
        return self["/attr/instrument"]

    @property
    def antenna_temperature(self):
        """Get antenna temperature."""
        file_key = self.filetype_info["antenna_temperature"]
        return self[file_key]

    @property
    def attrs(self):
        """Return attributes."""
        return {
            "filename": self.filename,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "platform_name": self.platform_name,
            "sensor": self.sensor,
        }

    @staticmethod
    def _standardize_dims(dataset):
        """Standardize dims to y, x."""
        if "atrack" in dataset.dims:
            dataset = dataset.rename({"atrack": "y"})
        if "xtrack" in dataset.dims:
            dataset = dataset.rename({"xtrack": "x"})
        if dataset.dims[0] == "x":
            dataset = dataset.transpose("y", "x")
        return dataset

    @staticmethod
    def _drop_coords(dataset):
        """Drop coords that are not in dims."""
        for coord in dataset.coords:
            if coord not in dataset.dims:
                dataset = dataset.drop_vars(coord)
        return dataset

    def _merge_attributes(self, dataset, dataset_info):
        """Merge attributes of the dataset."""
        dataset.attrs.update(self.filename_info)
        dataset.attrs.update(dataset_info)
        dataset.attrs.update(self.attrs)
        return dataset

    def _select_dataset(self, name):
        """Select dataset."""
        try:
            idx = int(name) - 1
            return self.antenna_temperature[:, :, idx]
        except ValueError:
            return self[name]

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset."""
        name = dataset_id["name"]
        logger.debug(f"Reading in file to get dataset with name {name}.")
        dataset = self._select_dataset(name)
        dataset = self._merge_attributes(dataset, ds_info)
        dataset = self._drop_coords(dataset)
        return self._standardize_dims(dataset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for GPM imerg data on half-hourly timesteps.

References:
   - The NASA IMERG ATBD:
     https://pmm.nasa.gov/sites/default/files/document_files/IMERG_ATBD_V06.pdf

"""

import datetime as dt
import logging

import dask.array as da
import h5py
import numpy as np
from pyresample.geometry import AreaDefinition

from satpy.readers.core.hdf5 import HDF5FileHandler

logger = logging.getLogger(__name__)


class Hdf5IMERG(HDF5FileHandler):
    """IMERG hdf5 reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init method."""
        super(Hdf5IMERG, self).__init__(filename, filename_info,
                                        filetype_info)
        self.finfo = filename_info
        self.cache = {}

    @property
    def start_time(self):
        """Find the start time from filename info."""
        return dt.datetime(self.finfo["date"].year,
                           self.finfo["date"].month,
                           self.finfo["date"].day,
                           self.finfo["start_time"].hour,
                           self.finfo["start_time"].minute,
                           self.finfo["start_time"].second)

    @property
    def end_time(self):
        """Find the end time from filename info."""
        return dt.datetime(self.finfo["date"].year,
                           self.finfo["date"].month,
                           self.finfo["date"].day,
                           self.finfo["end_time"].hour,
                           self.finfo["end_time"].minute,
                           self.finfo["end_time"].second)

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        file_key = ds_info.get("file_key", dataset_id["name"])
        dsname = "Grid/" + file_key
        data = self.get(dsname)
        data = data.squeeze().transpose()
        if data.ndim >= 2:
            data = data.rename({data.dims[-2]: "y", data.dims[-1]: "x"})
        data.data = da.flip(data.data, axis=0)

        fill = data.attrs["_FillValue"]
        data = data.where(data != fill)

        for key in list(data.attrs.keys()):
            val = data.attrs[key]
            if isinstance(val, h5py.h5r.Reference):
                del data.attrs[key]
            if isinstance(val, np.ndarray) and isinstance(val[0][0], h5py.h5r.Reference):
                del data.attrs[key]
        return data

    def get_area_def(self, dsid):
        """Create area definition from the gridded lat/lon values."""
        lats = self.__getitem__("Grid/lat").values
        lons = self.__getitem__("Grid/lon").values

        width = lons.shape[0]
        height = lats.shape[0]

        lower_left_x = lons[0]
        lower_left_y = lats[0]

        upper_right_x = lons[-1]
        upper_right_y = lats[-1]

        area_extent = (lower_left_x, lower_left_y, upper_right_x, upper_right_y)
        description = "IMERG GPM Equirectangular Projection"
        area_id = "imerg"
        proj_id = "equirectangular"
        proj_dict = {"proj": "longlat", "datum": "WGS84", "ellps": "WGS84", }
        area_def = AreaDefinition(area_id, description, proj_id, proj_dict, width, height, area_extent, )
        return area_def



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""CF encoding."""
import logging

import numpy as np
import xarray as xr
from xarray.coding.times import CFDatetimeCoder

logger = logging.getLogger(__name__)


def _set_default_chunks(encoding, dataset):
    """Update encoding to preserve current dask chunks.

    Existing user-defined chunks take precedence.
    """
    for var_name, variable in dataset.variables.items():
        if variable.chunks:
            chunks = tuple(
                np.stack([variable.data.chunksize,
                          variable.shape]).min(axis=0)
            )  # Chunksize may not exceed shape
            encoding.setdefault(var_name, {})
            encoding[var_name].setdefault("chunksizes", chunks)
    return encoding


def _set_default_fill_value(encoding, dataset):
    """Set default fill values.

    Avoid _FillValue attribute being added to coordinate variables
    (https://github.com/pydata/xarray/issues/1865).
    """
    coord_vars = []
    for data_array in dataset.values():
        coord_vars.extend(set(data_array.dims).intersection(data_array.coords))
    for coord_var in coord_vars:
        encoding.setdefault(coord_var, {})
        encoding[coord_var].update({"_FillValue": None})
    return encoding


def _set_default_time_encoding(encoding, dataset):
    """Set default time encoding.

    Make sure time coordinates and bounds have the same units.
    Default is xarray's CF datetime encoding, which can be overridden
    by user-defined encoding.
    """
    if "time" in dataset:
        try:
            dtnp64 = dataset["time"].data[0]
        except IndexError:
            dtnp64 = dataset["time"].data

        default = CFDatetimeCoder().encode(xr.DataArray(dtnp64))
        time_enc = {"units": default.attrs["units"], "calendar": default.attrs["calendar"]}
        time_enc.update(encoding.get("time", {}))
        bounds_enc = {"units": time_enc["units"],
                      "calendar": time_enc["calendar"],
                      "_FillValue": None}
        encoding["time"] = time_enc
        encoding["time_bnds"] = bounds_enc  # FUTURE: Not required anymore with xarray-0.14+
    return encoding


def _update_encoding_dataset_names(encoding, dataset, numeric_name_prefix):
    """Ensure variable names of the encoding dictionary account for numeric_name_prefix.

    A lot of channel names in satpy starts with a digit.
    When preparing CF-compliant datasets, these channels are prefixed with numeric_name_prefix.

    If variables names in the encoding dictionary are numeric digits, their name is prefixed
    with numeric_name_prefix
    """
    for var_name in list(dataset.variables):
        if not numeric_name_prefix or not var_name.startswith(numeric_name_prefix):
            continue
        orig_var_name = var_name.replace(numeric_name_prefix, "")
        if orig_var_name in encoding:
            encoding[var_name] = encoding.pop(orig_var_name)
    return encoding


def update_encoding(dataset, to_engine_kwargs, numeric_name_prefix="CHANNEL_"):
    """Update encoding.

    Preserve dask chunks, avoid fill values in coordinate variables and make sure that
    time & time bounds have the same units.
    """
    other_to_engine_kwargs = to_engine_kwargs.copy()
    encoding = other_to_engine_kwargs.pop("encoding", {}).copy()
    encoding = _update_encoding_dataset_names(encoding, dataset, numeric_name_prefix)
    encoding = _set_default_chunks(encoding, dataset)
    encoding = _set_default_fill_value(encoding, dataset)
    encoding = _set_default_time_encoding(encoding, dataset)
    return encoding, other_to_engine_kwargs



#!/usr/bin/env python
# Copyright (c) 2016-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests that compositors which were moved from package init to new modules raise warnings."""

import pytest


@pytest.mark.parametrize(
    "name",
    [
        "add_alpha_bands",
        "add_bands",
        "BackgroundCompositor",
        "CategoricalDataCompositor",
        "check_times",
        "CloudCompositor",
        "ColorizeCompositor",
        "ColormapCompositor",
        "CompositeBase",
        "DayNightCompositor",
        "DifferenceCompositor",
        "enhance2dataset",
        "Filler",
        "FillingCompositor",
        "GenericCompositor",
        "HighCloudCompositor",
        "IncompatibleAreas",
        "IncompatibleTimes",
        "LongitudeMaskingCompositor",
        "LowCloudCompositor",
        "LuminanceSharpeningCompositor",
        "MaskingCompositor",
        "MultiFiller",
        "NaturalEnh",
        "PaletteCompositor",
        "RatioCompositor",
        "RatioSharpenedRGB",
        "RealisticColors",
        "RGBCompositor",
        "SandwichCompositor",
        "SelfSharpenedRGB",
        "SimpleFireMaskCompositor",
        "SingleBandCompositor",
        "StaticImageCompositor",
        "SumCompositor",
        "zero_missing_data",
    ]
)
def test_import_from_package_init_warns(name):
    """Test that compositor classes and helper functions raise warnings when imported from package."""
    from satpy import composites
    with pytest.warns(UserWarning, match="has been moved to"):
        getattr(composites, name)



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the IASI NG L2 reader using temporary NetCDF files."""
import os
import re
from datetime import datetime

import h5py
import numpy as np
import pytest
import xarray as xr

from satpy import Scene
from satpy.readers.core.loading import load_reader
from satpy.readers.iasi_ng_l2_nc import IASINGL2NCFileHandler

d_lff = ("n_lines", "n_for", "n_fov")

DATA_DESC = [
    {
        "group": "data/diagnostics",
        "attrs": ["missing_value"],
        "variables": {
            "fg_cost": [d_lff, "float32", (3.4e38,)],
            "nbr_iterations": [d_lff, "int32", (2147483647,)],
            "rt_cost_x": [d_lff, "float32", (3.4e38,)],
            "rt_cost_y": [d_lff, "float32", (3.4e38,)],
        },
    },
    {
        "group": "data/geolocation_information",
        "attrs": [
            "units",
            "valid_min",
            "valid_max",
            "scale_factor",
            "add_offset",
            "missing_value",
        ],
        "variables": {
            "onboard_utc": [
                ("n_lines", "n_for"),
                "float64",
                ("seconds since 2020-01-01 00:00:00.000", -1e9, 1e9, None, None, -9e9),
            ],
            "sounder_pixel_latitude": [
                d_lff,
                "int32",
                ("degrees_north", -18e8, 18e8, 5.0e-8, 0.0, -2147483648),
            ],
            "sounder_pixel_longitude": [
                d_lff,
                "int32",
                ("degrees_east", -18.432e8, 18.432e8, 9.765625e-8, 0.0, -2147483648),
            ],
            "sounder_pixel_sun_azimuth": [
                d_lff,
                "int32",
                ("degrees", -18.432e8, 18.432e8, 9.765625e-8, 0.0, -2147483648),
            ],
            "sounder_pixel_sun_zenith": [
                d_lff,
                "int32",
                ("degrees", -18e8, 18e8, 5.0e-8, 90.0, -2147483648),
            ],
            "sounder_pixel_azimuth": [
                d_lff,
                "int32",
                ("degrees", -18.432e8, 18.432e8, 9.765625e-8, 0.0, -2147483648),
            ],
            "sounder_pixel_zenith": [
                d_lff,
                "int32",
                ("degrees", -18e8, 18e8, 2.5e-8, 45.0, -2147483648),
            ],
        },
    },
    {
        "group": "data/statistical_retrieval",
        "attrs": ["units", "valid_min", "valid_max", "missing_value"],
        "variables": {
            "air_temperature": [d_lff, "float32", ("K", 100.0, 400.0, 3.4e38)],
            "atmosphere_mass_content_of_water": [
                d_lff,
                "float32",
                ("kg.m-2", 0, 300, 3.4e38),
            ],
            "qi_air_temperature": [d_lff, "float32", ("", 0, 25, 3.4e38)],
            "qi_specific_humidity": [d_lff, "float32", ("", 0, 25, 3.4e38)],
            "specific_humidity": [d_lff, "float32", ("kg/kg", 0, 1, 3.4e38)],
            "surface_air_temperature": [d_lff, "float32", ("K", 100, 400, 3.4e38)],
            "surface_specific_humidity": [d_lff, "float32", ("K", 100, 400, 3.4e38)],
        },
    },
    {
        "group": "data/surface_info",
        "attrs": ["units", "valid_min", "valid_max", "missing_value"],
        "variables": {
            "height": [d_lff, "float32", ("m", -418, 8848, 3.4e38)],
            "height_std": [d_lff, "float32", ("m", 0, 999.0, 3.4e38)],
            "ice_fraction": [d_lff, "float32", ("%", 0, 100.0, 3.4e38)],
            "land_fraction": [d_lff, "float32", ("%", 0, 100.0, 3.4e38)],
        },
    },
]


def create_random_data(shape, dtype, attribs):
    """Create a random array with potentially missing values."""
    rng = np.random.default_rng()
    if dtype.startswith("float"):
        vmin = attribs.get("valid_min", -1e16)
        vmax = attribs.get("valid_max", 1e16)
        data = rng.uniform(vmin, vmax, shape).astype(dtype)
    else:
        vmin = attribs.get("valid_min", -2147483647)
        vmax = attribs.get("valid_max", 2147483647)
        data = rng.integers(vmin, vmax, shape, dtype=dtype)

    if "missing_value" in attribs:
        mask = rng.random(shape) < 0.05
        data[mask] = attribs["missing_value"]

    return data


def create_controlled_data(shape, dtype, attribs):
    """Create data with controlled values for testing scaling and masking."""
    data = np.zeros(shape, dtype=dtype)

    indices = np.indices(shape)
    rows, cols, fovs = indices

    if dtype.startswith("int"):
        valid_max = int(attribs.get("valid_max", 1000))
        data = ((rows * 100) + (cols * 10) + fovs) % valid_max
    else:
        valid_max = float(attribs.get("valid_max", 100.0))
        data = ((rows * 10.0) + (cols * 1.0) + (fovs / 10.0)) % valid_max

    if "missing_value" in attribs:
        data[0, 0, 0] = attribs["missing_value"]
        data[2, 2, 2] = attribs["missing_value"]

    return data


def get_valid_attribs(anames, alist):
    """Retrieve only the valid attributes from a list."""
    attribs = {}
    for idx, val in enumerate(alist):
        if val is not None:
            attribs[anames[idx]] = val

    return attribs


def add_dataset_variable(grp, var_config, dim_refs):
    """Add a variable to a given dataset group."""
    dnames = var_config["dims"]
    vname = var_config["name"]
    tname = var_config["dtype"]
    anames = var_config["allowed_attrs"]
    attribs = get_valid_attribs(anames, var_config["attrs"])

    shape = tuple([dim_refs[dname].size for dname in dnames])

    if vname in ["sounder_pixel_latitude", "sounder_pixel_longitude"]:
        arr = create_controlled_data(shape, tname, attribs)
    else:
        arr = create_random_data(shape, tname, attribs)

    dset = grp.create_dataset(vname, data=arr)

    for i, dname in enumerate(dnames):
        dset.dims[i].attach_scale(dim_refs[dname])

    for attr_name, attr_value in attribs.items():
        if attr_name in ["valid_min", "valid_max", "missing_value"]:
            attr_value = np.dtype(tname).type(attr_value)
        dset.attrs[attr_name] = attr_value


def write_fake_iasing_l2_file(output_path):
    """Create a fake IASI-NG L2 file with all required variables and attributes."""
    n_lines = 10
    n_for = 14
    n_fov = 16
    dims = {"n_lines": n_lines, "n_for": n_for, "n_fov": n_fov}

    with h5py.File(output_path, "w") as ds:
        grp = ds.create_group("data")
        dim_refs = {}
        for dim_name, size in dims.items():
            dim_dset = grp.create_dataset(dim_name, (size,), dtype="int32")
            dim_dset.make_scale(dim_name)
            dim_dset[:] = range(size)
            dim_refs[dim_name] = dim_dset

        for grp_desc in DATA_DESC:
            prefix = grp_desc["group"]
            grp = ds.create_group(prefix)

            for vname, vdesc in grp_desc["variables"].items():
                var_config = {
                    "name": vname,
                    "dims": vdesc[0],
                    "dtype": vdesc[1],
                    "attrs": vdesc[2],
                    "allowed_attrs": grp_desc["attrs"],
                }

                add_dataset_variable(grp, var_config, dim_refs)

    return os.fspath(output_path)


class TestIASINGL2NCReader:
    """Test class for the IASI NG L2 reader."""

    reader_name = "iasi_ng_l2_nc"
    file_prefix = "W_XX-EUMETSAT-Darmstadt,SAT,SGA1-IAS-02"
    file_suffix = "C_EUMT_20170616120000_G_V_20070912084329_20070912084600_O_N____.nc"

    def setup_method(self):
        """Set up the reader configuration."""
        from satpy._config import config_search_paths

        self.reader_configs = config_search_paths(
            os.path.join("readers", self.reader_name + ".yaml")
        )

    @pytest.fixture
    def twv_handler(self, tmp_path):
        """Create a simple default handler on a TWV product."""
        twv_filename = f"{self.file_prefix}-TWV_{self.file_suffix}"
        output_path = write_fake_iasing_l2_file(tmp_path / twv_filename)

        return self._create_file_handler(output_path)

    @pytest.fixture
    def twv_scene(self, tmp_path):
        """Create a simple satpy scene on a TWV product."""
        twv_filename = f"{self.file_prefix}-TWV_{self.file_suffix}"
        output_path = write_fake_iasing_l2_file(tmp_path / twv_filename)
        return Scene(filenames=[output_path], reader=self.reader_name)

    def _create_file_handler(self, filename):
        """Create an handler for the given file checking that it can be parsed correctly."""
        reader = load_reader(self.reader_configs)

        files = reader.select_files_from_pathnames([filename])
        assert len(files) == 1, "File should be recognized by the reader"

        reader.create_filehandlers(files)

        assert len(reader.file_handlers) == 1

        assert self.reader_name in reader.file_handlers

        handlers = reader.file_handlers[self.reader_name]

        assert len(handlers) == 1
        assert isinstance(handlers[0], IASINGL2NCFileHandler)

        return handlers[0]

    def test_filename_matching(self, tmp_path):
        """Test filename matching against some random name."""
        prefix = "W_fr-meteo-sat,GRAL,MTI1-IASING-2"
        suffix = (
            "C_EUMS_20220101120000_LEO_O_D_20220101115425_20220101115728_____W______.nc"
        )
        filename = f"{prefix}-l2p_{suffix}"
        output_path = write_fake_iasing_l2_file(tmp_path / filename)

        self._create_file_handler(output_path)

    def test_real_filename_matching(self, tmp_path):
        """Test that we will match an actual IASI NG L2 product file name."""
        supported_types = ["TWV", "CLD", "GHG", "SFC", "O3_", "CO_"]

        for ptype in supported_types:
            filename = f"{self.file_prefix}-{ptype}_{self.file_suffix}"
            output_path = write_fake_iasing_l2_file(tmp_path / filename)
            handler = self._create_file_handler(output_path)

            assert handler.filename_info["oflag"] == "C"
            assert handler.filename_info["originator"] == "EUMT"
            assert handler.filename_info["product_type"] == ptype

    def test_sensing_times(self, twv_handler):
        """Test that we read the sensing start/end times correctly from filename."""
        assert twv_handler.start_time == datetime(2007, 9, 12, 8, 43, 29)
        assert twv_handler.end_time == datetime(2007, 9, 12, 8, 46, 0)

    def test_sensor_names(self, twv_handler):
        """Test that the handler reports iasi_ng as sensor."""
        assert twv_handler.sensor_names == {"iasi_ng"}

    def test_available_datasets(self, twv_scene):
        """Test the list of available datasets in scene."""
        dnames = twv_scene.available_dataset_names()

        expected_names = [
            "onboard_utc",
            "sounder_pixel_latitude",
            "sounder_pixel_longitude",
            "sounder_pixel_azimuth",
            "sounder_pixel_zenith",
            "sounder_pixel_sun_azimuth",
            "sounder_pixel_sun_zenith",
            "air_temperature",
            "atmosphere_mass_content_of_water",
            "qi_air_temperature",
            "qi_specific_humidity",
            "specific_humidity",
            "surface_air_temperature",
            "surface_specific_humidity",
        ]

        for dname in expected_names:
            assert dname in dnames

    def test_latlon_datasets(self, twv_scene):
        """Test loading the latitude/longitude dataset."""
        twv_scene.load(["sounder_pixel_latitude", "sounder_pixel_longitude"])
        lat = twv_scene["sounder_pixel_latitude"]
        lon = twv_scene["sounder_pixel_longitude"]

        assert lat.dims == ("n_lines", "n_for", "n_fov")
        assert lon.dims == ("n_lines", "n_for", "n_fov")

        assert lat.dtype == np.float64
        assert lon.dtype == np.float64

        assert np.nanmin(lat) >= -90.0
        assert np.nanmax(lat) <= 90.0

        assert np.nanmin(lon) >= -180.0
        assert np.nanmax(lon) <= 180.0

    def test_scaling_latitude(self, tmp_path):
        """Test that latitude values are correctly scaled and masked."""
        twv_filename = f"{self.file_prefix}-TWV_{self.file_suffix}"
        output_path = write_fake_iasing_l2_file(tmp_path / twv_filename)

        scene = Scene(filenames=[output_path], reader=self.reader_name)
        scene.load(["sounder_pixel_latitude"])
        lat = scene["sounder_pixel_latitude"]

        with h5py.File(output_path, "r") as f:
            raw_lat = f["data/geolocation_information/sounder_pixel_latitude"][:]
            scale_factor = f["data/geolocation_information/sounder_pixel_latitude"].attrs[
                "scale_factor"
            ]
            add_offset = f["data/geolocation_information/sounder_pixel_latitude"].attrs[
                "add_offset"
            ]
            missing_value = f[
                "data/geolocation_information/sounder_pixel_latitude"
            ].attrs["missing_value"]

        assert np.isnan(lat.values[0, 0, 0])

        valid_point = np.where(raw_lat != missing_value)
        if len(valid_point[0]) > 0:
            i, j, k = valid_point[0][0], valid_point[1][0], valid_point[2][0]
            expected_value = raw_lat[i, j, k] * scale_factor + add_offset
            np.testing.assert_almost_equal(lat.values[i, j, k], expected_value)

        assert lat.attrs["units"] == "degrees_north"

    def test_mask_propagation(self, twv_scene):
        """Test that masks are properly propagated to derived datasets."""
        twv_scene.load(["sounder_pixel_latitude", "sounder_pixel_longitude"])
        lat = twv_scene["sounder_pixel_latitude"]
        lon = twv_scene["sounder_pixel_longitude"]

        assert np.isnan(lat.values[0, 0, 0])
        assert np.isnan(lon.values[0, 0, 0])

        assert np.isnan(lat.values[2, 2, 2])
        assert np.isnan(lon.values[2, 2, 2])

    def test_onboard_utc_dataset(self, twv_scene):
        """Test loading the onboard_utc dataset."""
        twv_scene.load(["onboard_utc", "sounder_pixel_latitude"])
        dset = twv_scene["onboard_utc"]

        assert dset.dims == ("n_lines", "n_for")

        assert dset.dtype == np.dtype("datetime64[ns]")

    def test_nbr_iterations_dataset(self, twv_scene):
        """Test loading the nbr_iterations dataset."""
        twv_scene.load(["nbr_iterations"])
        dset = twv_scene["nbr_iterations"]

        assert len(dset.dims) == 3
        assert dset.dtype == np.float64

    def test_register_dataset(self, twv_handler):
        """Test the register_dataset method."""
        twv_handler.dataset_infos = {}
        twv_handler.register_dataset("test_dataset", {"attr": "value"})
        assert "test_dataset" in twv_handler.dataset_infos
        assert twv_handler.dataset_infos["test_dataset"]["attr"] == "value"

        with pytest.raises(KeyError):
            twv_handler.register_dataset("test_dataset", {"attr": "new_value"})

    def test_process_dimension(self, twv_handler):
        """Test the process_dimension method."""
        twv_handler.dimensions_desc = {}
        twv_handler.process_dimension("group/dimension/test_dim", 10)
        assert twv_handler.dimensions_desc["test_dim"] == 10

        with pytest.raises(KeyError):
            twv_handler.process_dimension("group/dimension/test_dim", 20)

    def test_process_attribute(self, twv_handler):
        """Test the process_attribute method."""
        twv_handler.variable_desc = {"test_var": {"attribs": {}}}
        twv_handler.process_attribute("test_var/attr/test_attr", "value")
        assert twv_handler.variable_desc["test_var"]["attribs"]["test_attr"] == "value"

    def test_process_variable(self, twv_handler):
        """Test the process_variable method."""
        twv_handler.file_content = {
            "group/test_var/shape": (10, 10),
            "group/test_var/dimensions": ("x", "y"),
            "group/test_var/dtype": "float32",
        }
        twv_handler.ignored_patterns = []
        twv_handler.process_variable("group/test_var")
        assert "group/test_var" in twv_handler.variable_desc

    def test_ignore_scalar_variable(self, twv_handler):
        """Test ignoring of scalar variable."""
        twv_handler.file_content = {
            "group/test_var/shape": (1, 1),
            "group/test_var/dimensions": ("x", "y"),
            "group/test_var/dtype": "float32",
        }
        twv_handler.ignored_patterns = []
        twv_handler.process_variable("group/test_var")
        assert "group/test_var" not in twv_handler.variable_desc

    def test_ignore_pattern_variable(self, twv_handler):
        """Test ignoring of pattern in variable."""
        twv_handler.file_content = {
            "group/test_var/shape": (10, 10),
            "group/test_var/dimensions": ("x", "y"),
            "group/test_var/dtype": "float32",
        }
        twv_handler.ignored_patterns = [re.compile(r"test_")]
        twv_handler.process_variable("group/test_var")
        assert "group/test_var" not in twv_handler.variable_desc

    def test_parse_file_content(self, twv_handler):
        """Test the parse_file_content method."""
        twv_handler.file_content = {
            "dim/dimension/test_dim": 10,
            "var/attr/test_attr": "value",
            "grp/test_var": "my_data",
            "grp/test_var/shape": (10, 10),
            "grp/test_var/dimensions": ("x", "y"),
            "grp/test_var/dtype": "float32",
        }

        twv_handler.parse_file_content()
        assert "test_dim" in twv_handler.dimensions_desc
        assert "grp/test_var" in twv_handler.variable_desc

    def test_register_available_datasets(self, twv_handler):
        """Test the register_available_datasets method."""
        twv_handler.dataset_infos = None
        twv_handler.variable_desc = {
            "var/test_var": {"var_name": "test_var", "attribs": {"units": "test_unit"}}
        }
        twv_handler.dataset_aliases = {}
        twv_handler.register_available_datasets()
        assert "test_var" in twv_handler.dataset_infos

    def test_ignored_register_available_datasets(self, twv_handler):
        """Test ignoring register_available_datasets method if done already."""
        twv_handler.variable_desc = {
            "var/test_var": {"var_name": "test_var", "attribs": {"units": "test_unit"}}
        }
        twv_handler.dataset_aliases = {}
        twv_handler.register_available_datasets()
        assert "test_var" not in twv_handler.dataset_infos

    def test_register_available_datasets_alias(self, twv_handler):
        """Test the register_available_datasets method with alias."""
        twv_handler.dataset_infos = None
        twv_handler.variable_desc = {
            "var/test_var": {"var_name": "test_var", "attribs": {"units": "test_unit"}}
        }
        twv_handler.dataset_aliases = {re.compile(r"var/(.+)$"): "${VAR_NAME}_oem"}
        twv_handler.register_available_datasets()
        assert "test_var_oem" in twv_handler.dataset_infos

    def test_get_dataset_infos(self, twv_handler):
        """Test the get_dataset_infos method."""
        twv_handler.dataset_infos = {"test_dataset": {"attr": "value"}}
        assert twv_handler.get_dataset_infos("test_dataset") == {"attr": "value"}

        with pytest.raises(KeyError):
            twv_handler.get_dataset_infos("non_existent_dataset")

    def test_variable_path_exists(self, twv_handler):
        """Test the variable_path_exists method."""
        twv_handler.file_content = {"test_var": "dummy"}
        assert twv_handler.variable_path_exists("test_var")
        assert not twv_handler.variable_path_exists("/attr/test_var")
        assert not twv_handler.variable_path_exists("test_var/dtype")
        assert not twv_handler.variable_path_exists("/grp/a_non_existing_var")

    def test_convert_to_datetime(self, twv_handler):
        """Test the convert_to_datetime method."""
        data = xr.DataArray(np.array([0, 86400]), dims=("time",))
        ds_info = {"seconds_since_epoch": "2000-01-01 00:00:00"}
        result = twv_handler.convert_to_datetime(data, ds_info)
        assert result[0].values == np.datetime64("2000-01-01T00:00:00")
        assert result[1].values == np.datetime64("2000-01-02T00:00:00")

    def test_get_transformed_dataset(self, twv_handler):
        """Test the get_transformed_dataset method."""
        ds_info = {
            "location": "test_var",
            "seconds_since_epoch": "2000-01-01 00:00:00",
        }
        twv_handler.variable_path_exists = lambda _arg: True
        twv_handler.file_content["test_var"] = xr.DataArray(
            np.array([[0, 86400]]), dims=("x", "y")
        )

        twv_handler.dimensions_desc = {"n_fov": 2}

        result = twv_handler.get_transformed_dataset(ds_info)
        assert result.shape == (1, 2)
        assert result.dtype == "datetime64[ns]"

    def test_get_transformed_dataset_failure(self, twv_handler):
        """Test the get_transformed_dataset method fails on invalid path."""
        ds_info = {
            "location": "test_var",
            "seconds_since_epoch": "2000-01-01 00:00:00",
        }

        with pytest.raises(KeyError):
            twv_handler.get_transformed_dataset(ds_info)

    def test_get_dataset(self, twv_handler):
        """Test the get_dataset method."""
        twv_handler.dataset_infos = {
            "test_dataset": {"name": "test_dataset", "location": "test_var"}
        }
        twv_handler.get_transformed_dataset = lambda _arg: xr.DataArray([1, 2, 3])

        result = twv_handler.get_dataset({"name": "test_dataset"})
        assert result.equals(xr.DataArray([1, 2, 3]))

        assert twv_handler.get_dataset({"name": "non_existent_dataset"}) is None



"""Helpers for loading writer config files."""
from __future__ import annotations

import logging
import os

import yaml
from yaml import UnsafeLoader

from satpy._config import config_search_paths, get_entry_points_config_dirs, glob_config

LOG = logging.getLogger(__name__)


def read_writer_config(config_files, loader=UnsafeLoader):
    """Read the writer `config_files` and return the info extracted."""
    conf = {}
    LOG.debug("Reading %s", str(config_files))
    for config_file in config_files:
        with open(config_file) as fd:
            conf.update(yaml.load(fd.read(), Loader=loader))

    try:
        writer_info = conf["writer"]
    except KeyError:
        raise KeyError(
            "Malformed config file {}: missing writer 'writer'".format(
                config_files))
    writer_info["config_files"] = config_files
    return writer_info


def load_writer_configs(writer_configs, **writer_kwargs):
    """Load the writer from the provided `writer_configs`."""
    try:
        writer_info = read_writer_config(writer_configs)
        writer_class = writer_info["writer"]
    except (ValueError, KeyError, yaml.YAMLError):
        raise ValueError("Invalid writer configs: "
                         "'{}'".format(writer_configs))
    init_kwargs, kwargs = writer_class.separate_init_kwargs(writer_kwargs)
    writer = writer_class(config_files=writer_configs,
                          **init_kwargs)
    return writer, kwargs


def load_writer(writer, **writer_kwargs):
    """Find and load writer `writer` in the available configuration files."""
    config_fn = writer + ".yaml" if "." not in writer else writer
    config_files = config_search_paths(os.path.join("writers", config_fn))
    writer_kwargs.setdefault("config_files", config_files)
    if not writer_kwargs["config_files"]:
        raise ValueError("Unknown writer '{}'".format(writer))

    try:
        return load_writer_configs(writer_kwargs["config_files"],
                                   **writer_kwargs)
    except ValueError:
        raise ValueError("Writer '{}' does not exist or could not be "
                         "loaded".format(writer))


def configs_for_writer(writer=None):
    """Generate writer configuration files for one or more writers.

    Args:
        writer (Optional[str]): Yield configs only for this writer

    Returns: Generator of lists of configuration files

    """
    if writer is not None:
        if not isinstance(writer, (list, tuple)):
            writer = [writer]
        # given a config filename or writer name
        config_files = [w if w.endswith(".yaml") else w + ".yaml" for w in writer]
    else:
        paths = get_entry_points_config_dirs("satpy.writers")
        writer_configs = glob_config(os.path.join("writers", "*.yaml"), search_dirs=paths)
        config_files = set(writer_configs)

    for config_file in config_files:
        config_basename = os.path.basename(config_file)
        paths = get_entry_points_config_dirs("satpy.writers")
        writer_configs = config_search_paths(
            os.path.join("writers", config_basename),
            search_dirs=paths,
        )

        if not writer_configs:
            LOG.warning("No writer configs found for '%s'", writer)
            continue

        yield writer_configs


def available_writers(as_dict=False):
    """Available writers based on current configuration.

    Args:
        as_dict (bool): Optionally return writer information as a dictionary.
                        Default: False

    Returns: List of available writer names. If `as_dict` is `True` then
             a list of dictionaries including additionally writer information
             is returned.

    """
    writers = []
    for writer_configs in configs_for_writer():
        try:
            writer_info = read_writer_config(writer_configs)
        except (KeyError, IOError, yaml.YAMLError):
            LOG.warning("Could not import writer config from: %s", writer_configs)
            LOG.debug("Error loading YAML", exc_info=True)
            continue
        writers.append(writer_info if as_dict else writer_info["name"])
    return writers



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for eps level 1b data. Uses xml files as a format description."""

import functools
import logging

import dask.array as da
import numpy as np
import xarray as xr
from dask.delayed import delayed
from pyresample.geometry import SwathDefinition

from satpy._compat import cached_property
from satpy._config import get_config_path
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.xmlformat import XMLFormat
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

C1 = 1.191062e-05  # mW/(m2*sr*cm-4)
C2 = 1.4387863  # K/cm-1


def radiance_to_bt(arr, wc_, a__, b__):
    """Convert to BT in K."""
    return a__ + b__ * (C2 * wc_ / (da.log(1 + (C1 * (wc_ ** 3) / arr))))


def radiance_to_refl(arr, solar_flux):
    """Convert to reflectances in %."""
    return arr * np.pi * 100.0 / solar_flux


record_class = ["Reserved", "mphr", "sphr",
                "ipr", "geadr", "giadr",
                "veadr", "viadr", "mdr"]


def read_records(filename):
    """Read *filename* without scaling it afterwards."""
    format_fn = get_config_path("eps_avhrrl1b_6.5.xml")
    form = XMLFormat(format_fn)

    grh_dtype = np.dtype([("record_class", "|i1"),
                          ("INSTRUMENT_GROUP", "|i1"),
                          ("RECORD_SUBCLASS", "|i1"),
                          ("RECORD_SUBCLASS_VERSION", "|i1"),
                          ("RECORD_SIZE", ">u4"),
                          ("RECORD_START_TIME", "S6"),
                          ("RECORD_STOP_TIME", "S6")])

    max_lines = np.floor((CHUNK_SIZE ** 2) / 2048)

    dtypes = []
    cnt = 0
    counts = []
    classes = []
    prev = None
    with open(filename, "rb") as fdes:
        while True:
            grh = np.fromfile(fdes, grh_dtype, 1)
            if grh.size == 0:
                break
            rec_class = record_class[int(grh["record_class"].squeeze())]
            sub_class = grh["RECORD_SUBCLASS"][0]

            expected_size = int(grh["RECORD_SIZE"].squeeze())
            bare_size = expected_size - grh_dtype.itemsize
            try:
                the_type = form.dtype((rec_class, sub_class))
                # the_descr = grh_dtype.descr + the_type.descr
            except KeyError:
                the_type = np.dtype([("unknown", "V%d" % bare_size)])
            the_descr = grh_dtype.descr + the_type.descr
            the_type = np.dtype(the_descr)
            if the_type.itemsize < expected_size:
                padding = [("unknown%d" % cnt, "V%d" % (expected_size - the_type.itemsize))]
                cnt += 1
                the_descr += padding
            new_dtype = np.dtype(the_descr)
            key = (rec_class, sub_class)
            if key == prev:
                counts[-1] += 1
            else:
                dtypes.append(new_dtype)
                counts.append(1)
                classes.append(key)
                prev = key
            fdes.seek(expected_size - grh_dtype.itemsize, 1)

        sections = {}
        offset = 0
        for dtype, count, rec_class in zip(dtypes, counts, classes):
            fdes.seek(offset)
            if rec_class == ("mdr", 2):
                record = da.from_array(np.memmap(fdes, mode="r", dtype=dtype, shape=count, offset=offset),
                                       chunks=(max_lines,))
            else:
                record = np.fromfile(fdes, dtype=dtype, count=count)
            offset += dtype.itemsize * count
            if rec_class in sections:
                logger.debug("Multiple records for ", str(rec_class))
                sections[rec_class] = np.hstack((sections[rec_class], record))
            else:
                sections[rec_class] = record

    return sections, form


def create_xarray(arr):
    """Create xarray with correct dimensions."""
    res = arr
    res = xr.DataArray(res, dims=["y", "x"])
    return res


class EPSAVHRRFile(BaseFileHandler):
    """Eps level 1b reader for AVHRR data."""

    spacecrafts = {"M01": "Metop-B",
                   "M02": "Metop-A",
                   "M03": "Metop-C", }

    sensors = {"AVHR": "avhrr-3"}

    units = {"reflectance": "%",
             "brightness_temperature": "K",
             "radiance":  "W m^-2 sr^-1"}

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize FileHandler."""
        super(EPSAVHRRFile, self).__init__(
            filename, filename_info, filetype_info)

        self.area = None
        self._start_time = filename_info["start_time"]
        self._end_time = filename_info["end_time"]
        self.form = None
        self.scanlines = None
        self.pixels = None
        self.sections = None
        self.get_full_angles = functools.lru_cache(maxsize=1)(
            self._get_full_angles_uncached
        )
        self.get_full_lonlats = functools.lru_cache(maxsize=1)(
            self._get_full_lonlats_uncached
        )

    def _read_all(self):
        logger.debug("Reading %s", self.filename)
        self.sections, self.form = read_records(self.filename)
        self.scanlines = self["TOTAL_MDR"]
        if self.scanlines != len(self.sections[("mdr", 2)]):
            logger.warning("Number of declared records doesn't match number of scanlines in the file.")
            self.scanlines = len(self.sections[("mdr", 2)])
        self.pixels = self["EARTH_VIEWS_PER_SCANLINE"]

    def __getitem__(self, key):
        """Get value for given key."""
        for altkey in self.form.scales:
            try:
                try:
                    return self.sections[altkey][key] * self.form.scales[altkey][key]
                except TypeError:
                    val = self.sections[altkey][key].item().decode().split("=")[1]
                    try:
                        return float(val) * self.form.scales[altkey][key].item()
                    except ValueError:
                        return val.strip()
            except (KeyError, ValueError):
                continue
        raise KeyError("No matching value for " + str(key))

    def keys(self):
        """List of reader's keys."""
        keys = []
        for val in self.form.scales.values():
            keys += val.dtype.fields.keys()
        return keys

    def _get_full_lonlats_uncached(self):
        """Get the interpolated longitudes and latitudes."""
        raw_lats = np.hstack((self["EARTH_LOCATION_FIRST"][:, [0]],
                              self["EARTH_LOCATIONS"][:, :, 0],
                              self["EARTH_LOCATION_LAST"][:, [0]]))

        raw_lons = np.hstack((self["EARTH_LOCATION_FIRST"][:, [1]],
                              self["EARTH_LOCATIONS"][:, :, 1],
                              self["EARTH_LOCATION_LAST"][:, [1]]))
        return self._interpolate(raw_lons, raw_lats)

    def _interpolate(self, lons_like, lats_like):
        nav_sample_rate = self["NAV_SAMPLE_RATE"]
        if nav_sample_rate == 20 and self.pixels == 2048:
            lons_like_1km, lats_like_1km = _interpolate_20km_to_1km(lons_like, lats_like)
            lons_like_1km = da.from_delayed(lons_like_1km, dtype=lons_like.dtype,
                                            shape=(self.scanlines, self.pixels))
            lats_like_1km = da.from_delayed(lats_like_1km, dtype=lats_like.dtype,
                                            shape=(self.scanlines, self.pixels))
            return lons_like_1km, lats_like_1km

        raise NotImplementedError("Lon/lat and angle expansion not implemented for " +
                                  "sample rate = " + str(nav_sample_rate) +
                                  " and earth views = " +
                                  str(self.pixels))

    def _get_full_angles(self, solar_zenith, sat_zenith, solar_azimuth, sat_azimuth):

        nav_sample_rate = self["NAV_SAMPLE_RATE"]
        if nav_sample_rate == 20 and self.pixels == 2048:
            # Note: interpolation assumes second array values between -90 and 90
            # Solar and satellite zenith is between 0 and 180.
            sun_azi, sun_zen = self._interpolate(solar_azimuth, solar_zenith - 90)
            sun_zen += 90
            sat_azi, sat_zen = self._interpolate(sat_azimuth, sat_zenith - 90)
            sat_zen += 90
            return sun_azi, sun_zen, sat_azi, sat_zen
        else:
            raise NotImplementedError("Angles expansion not implemented for " +
                                      "sample rate = " + str(nav_sample_rate) +
                                      " and earth views = " +
                                      str(self.pixels))

    def _get_full_angles_uncached(self):
        """Get the interpolated angles."""
        solar_zenith = np.hstack((self["ANGULAR_RELATIONS_FIRST"][:, [0]],
                                  self["ANGULAR_RELATIONS"][:, :, 0],
                                  self["ANGULAR_RELATIONS_LAST"][:, [0]]))

        sat_zenith = np.hstack((self["ANGULAR_RELATIONS_FIRST"][:, [1]],
                                self["ANGULAR_RELATIONS"][:, :, 1],
                                self["ANGULAR_RELATIONS_LAST"][:, [1]]))

        solar_azimuth = np.hstack((self["ANGULAR_RELATIONS_FIRST"][:, [2]],
                                   self["ANGULAR_RELATIONS"][:, :, 2],
                                   self["ANGULAR_RELATIONS_LAST"][:, [2]]))
        sat_azimuth = np.hstack((self["ANGULAR_RELATIONS_FIRST"][:, [3]],
                                 self["ANGULAR_RELATIONS"][:, :, 3],
                                 self["ANGULAR_RELATIONS_LAST"][:, [3]]))

        return self._get_full_angles(solar_zenith,
                                     sat_zenith,
                                     solar_azimuth,
                                     sat_azimuth)

    def get_bounding_box(self):
        """Get bounding box."""
        if self.sections is None:
            self._read_all()
        lats = np.hstack([self["EARTH_LOCATION_FIRST"][0, [0]],
                          self["EARTH_LOCATION_LAST"][0, [0]],
                          self["EARTH_LOCATION_LAST"][-1, [0]],
                          self["EARTH_LOCATION_FIRST"][-1, [0]]])
        lons = np.hstack([self["EARTH_LOCATION_FIRST"][0, [1]],
                          self["EARTH_LOCATION_LAST"][0, [1]],
                          self["EARTH_LOCATION_LAST"][-1, [1]],
                          self["EARTH_LOCATION_FIRST"][-1, [1]]])
        return lons.ravel(), lats.ravel()

    def get_dataset(self, key, info):
        """Get calibrated channel data."""
        if self.sections is None:
            self._read_all()

        try:
            dataset = self._get_data_array(key)
        except KeyError:
            logger.info("Can't load channel in eps_l1b: " + str(key["name"]))
            return

        dataset.attrs["platform_name"] = self.platform_name
        dataset.attrs["sensor"] = self.sensor_name
        if "calibration" in key:
            dataset.attrs["units"] = self.units[key["calibration"]]
        dataset.attrs.update(info)
        dataset.attrs.update(key.to_dict())
        return dataset

    def _get_data_array(self, key):
        name = key["name"]
        if name in ["longitude", "latitude"]:
            data = self.get_full_lonlats()[int(name == "latitude")]
            dataset = create_xarray(data)
        elif name in ["solar_zenith_angle", "solar_azimuth_angle", "satellite_zenith_angle", "satellite_azimuth_angle"]:
            dataset = self._get_angle_dataarray(key)
        elif name in ["1", "2", "3a", "3A", "3b", "3B", "4", "5"]:
            dataset = self._get_calibrated_dataarray(key)
        elif name == "cloud_flags":
            array = self["CLOUD_INFORMATION"]
            dataset = create_xarray(array)
        else:
            raise KeyError(f"Unknown channel: {name}")
        return dataset

    def _get_angle_dataarray(self, key):
        """Get an angle dataarray."""
        arr_index = {
            "solar_azimuth_angle": 0,
            "solar_zenith_angle": 1,
            "satellite_azimuth_angle": 2,
            "satellite_zenith_angle": 3,
        }[key["name"]]
        data = self.get_full_angles()[arr_index]
        return create_xarray(data)

    @cached_property
    def three_a_mask(self):
        """Mask for 3A."""
        return (self["FRAME_INDICATOR"] & 2 ** 16) != 2 ** 16

    @cached_property
    def three_b_mask(self):
        """Mask for 3B."""
        return (self["FRAME_INDICATOR"] & 2 ** 16) != 0

    def _get_calibrated_dataarray(self, key):
        """Get a calibrated dataarray."""
        if key["calibration"] not in ["reflectance", "brightness_temperature", "radiance"]:
            raise ValueError("calibration type " + str(key["calibration"]) +
                             " is not supported!")

        mask = None

        channel_name = key["name"].upper()

        radiance_indices = {"1": 0, "2": 1, "3A": 2, "3B": 2, "4": 3, "5": 4}
        array = self["SCENE_RADIANCES"][:, radiance_indices[channel_name], :]

        if channel_name in ["1", "2", "3A"]:
            if key["calibration"] == "reflectance":
                array = radiance_to_refl(array,
                                         self[f"CH{channel_name}_SOLAR_FILTERED_IRRADIANCE"])
            if channel_name == "3A":
                mask = self.three_a_mask[:, np.newaxis]

        if channel_name in ["3B", "4", "5"]:
            if key["calibration"] == "brightness_temperature":
                array = radiance_to_bt(array,
                                       self[f"CH{channel_name}_CENTRAL_WAVENUMBER"],
                                       self[f"CH{channel_name}_CONSTANT1"],
                                       self[f"CH{channel_name}_CONSTANT2_SLOPE"])
            if channel_name == "3B":
                mask = self.three_b_mask[:, np.newaxis]

        dataset = create_xarray(array)
        if mask is not None:
            dataset = dataset.where(~mask)
        return dataset

    def get_lonlats(self):
        """Get lonlats."""
        if self.area is None:
            lons, lats = self.get_full_lonlats()
            self.area = SwathDefinition(lons, lats)
            self.area.name = "_".join([self.platform_name, str(self.start_time),
                                       str(self.end_time)])
        return self.area

    @property
    def platform_name(self):
        """Get platform name."""
        return self.spacecrafts[self["SPACECRAFT_ID"]]

    @property
    def sensor_name(self):
        """Get sensor name."""
        return self.sensors[self["INSTRUMENT_ID"]]

    @property
    def start_time(self):
        """Get start time."""
        # return datetime.strptime(self["SENSING_START"], "%Y%m%d%H%M%SZ")
        return self._start_time

    @property
    def end_time(self):
        """Get end time."""
        # return datetime.strptime(self["SENSING_END"], "%Y%m%d%H%M%SZ")
        return self._end_time


@delayed(nout=2, pure=True)
def _interpolate_20km_to_1km(lons, lats):
    # Note: delayed will cast input dask-arrays to numpy arrays (needed by metop20kmto1km).
    from geotiepoints import metop20kmto1km
    return metop20kmto1km(lons, lats)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2024 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""MCD12Q1 hdf-eos format reader.

Introduction
------------

The ``mcd12q1`` reader reads MCD12Q1 products in HDF-EOS format.

The 500m product is provided  on a sinusoidal grid.

Reference documents and links:
    - MODIS land products grid: https://modis-land.gsfc.nasa.gov/MODLAND_grid.html
    - User guide: https://lpdaac.usgs.gov/documents/101/MCD12_User_Guide_V6.pdf
    - MCD12Q1 v061: MODIS/Terra+Aqua Land Cover Type Yearly L3 Global 500 m SIN Grid

The reader has been tested with:
    - MCD12Q1: Land cover data.

To get a list of the available datasets for a given file refer to the "Load data" section in :doc:`../reading`.

"""
import logging
from typing import Iterable

from pyresample import geometry

from satpy.readers.core.hdfeos import HDFEOSBaseFileReader

logger = logging.getLogger(__name__)


class MCD12Q1HDFFileHandler(HDFEOSBaseFileReader):
    """File handler for MCD12Q1 HDF-EOS 500m granules."""
    def available_datasets(self, configured_datasets=None):
        """Automatically determine datasets provided by this file."""
        # Initialise set of variable names to carry through code
        handled_var_names = set()

        ds_dict = self.sd.datasets()

        for is_avail, ds_info in (configured_datasets or []):
            file_key = ds_info.get("file_key", ds_info["name"])
            # we must add all variables here even if another file handler has
            # claimed the variable. It could be another instance of this file
            # type, and we don't want to add that variable dynamically if the
            # other file handler defined it by the YAML definition.
            handled_var_names.add(file_key)
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            if self.file_type_matches(ds_info["file_type"]) is None:
                # this is not the file type for this dataset
                yield None, ds_info
                continue
            yield file_key in ds_dict.keys(), ds_info

        yield from self._dynamic_variables_from_file(handled_var_names)

    def _dynamic_variables_from_file(self, handled_var_names: set) -> Iterable[tuple[bool, dict]]:
        res = self._get_res()
        for var_name in self.sd.datasets().keys():
            if var_name in handled_var_names:
                # skip variables that YAML had configured
                continue
            common = {"file_type": "mcd12q1_500m_hdf",
                      "resolution": res,
                      "name": var_name}
            yield True, common


    def _get_res(self):
        """Compute the resolution from the file metadata."""
        gridname = self.metadata["GridStructure"]["GRID_1"]["GridName"]
        if "MCD12Q1" not in gridname:
            raise ValueError("Only MCD12Q1 grids are supported")

        resolution_string = self.metadata["ARCHIVEDMETADATA"]["NADIRDATARESOLUTION"]["VALUE"]
        if resolution_string[-1] == "m":
            return int(resolution_string.removesuffix("m"))
        else:
            raise ValueError("Cannot parse resolution of MCD12Q1 grid")


    def get_dataset(self, dataset_id, dataset_info):
        """Get DataArray for specified dataset."""
        dataset_name = dataset_id["name"]
        # xxx
        dataset = self.load_dataset(dataset_name, dataset_info.pop("category", False))

        self._add_satpy_metadata(dataset_id, dataset)

        return dataset

    def _get_area_extent(self):
        """Get the grid properties."""
        # Now compute the data extent
        upperleft = self.metadata["GridStructure"]["GRID_1"]["UpperLeftPointMtrs"]
        lowerright = self.metadata["GridStructure"]["GRID_1"]["LowerRightMtrs"]

        return upperleft[0], lowerright[1], lowerright[0], upperleft[1]

    def get_area_def(self, dsid):
        """Get the area definition.

        This is fixed, but not defined in the file. So we must
        generate it ourselves with some assumptions.

        The proj_param string comes from https://lpdaac.usgs.gov/documents/101/MCD12_User_Guide_V6.pdf
        """
        proj_param = "proj=sinu +a=6371007.181 +b=6371007.181 +units=m"

        # Get the size of the dataset
        nrows = self.metadata["GridStructure"]["GRID_1"]["YDim"]
        ncols = self.metadata["GridStructure"]["GRID_1"]["XDim"]

        # Construct the area definition
        area = geometry.AreaDefinition("sinusoidal_modis",
                                       "Tiled sinusoidal L3 MODIS area",
                                       "sinusoidal",
                                       proj_param,
                                       ncols,
                                       nrows,
                                       self._get_area_extent())

        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
r"""Reader for files produced with the cf netcdf writer in satpy.

Introduction
------------

The ``satpy_cf_nc`` reader reads data written by the satpy cf_writer. Filenames for cf_writer are optional.
There are several readers using the same satpy_cf_nc.py reader.

* Generic reader ``satpy_cf_nc``
* EUMETSAT GAC FDR reader ``avhrr_l1c_eum_gac_fdr_nc``

Generic reader
--------------

The generic ``satpy_cf_nc`` reader reads files of type:

.. code-block:: none

    '{platform_name}-{sensor}-{start_time:%Y%m%d%H%M%S}-{end_time:%Y%m%d%H%M%S}.nc'


Example:
--------
Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene

    filenames = ['data/npp-viirs-mband-20201007075915-20201007080744.nc']
    scn = Scene(reader='satpy_cf_nc', filenames=filenames)
    scn.load(['M05'])
    scn['M05']

Output:

.. code-block:: none

    <xarray.DataArray 'M05' (y: 4592, x: 3200)>
    dask.array<open_dataset-d91cfbf1bf4f14710d27446d91cdc6e4M05, shape=(4592, 3200),
        dtype=float32, chunksize=(4096, 3200), chunktype=numpy.ndarray>
    Coordinates:
        longitude  (y, x) float32 dask.array<chunksize=(4096, 3200), meta=np.ndarray>
        latitude   (y, x) float32 dask.array<chunksize=(4096, 3200), meta=np.ndarray>
    Dimensions without coordinates: y, x
    Attributes:
        start_time:                   2020-10-07 07:59:15
        start_orbit:                  46350
        end_time:                     2020-10-07 08:07:44
        end_orbit:                    46350
        calibration:                  reflectance
        long_name:                    M05
        modifiers:                    ('sunz_corrected',)
        platform_name:                Suomi-NPP
        resolution:                   742
        sensor:                       viirs
        standard_name:                toa_bidirectional_reflectance
        units:                        %
        wavelength:                   0.672 µm (0.662-0.682 µm)
        date_created:                 2020-10-07T08:20:02Z
        instrument:                   VIIRS


Notes:
    Available datasets and attributes will depend on the data saved with the cf_writer.

EUMETSAT AVHRR GAC FDR L1C reader
---------------------------------

The ``avhrr_l1c_eum_gac_fdr_nc`` reader reads files of type:

.. code-block:: none

    ''AVHRR-GAC_FDR_1C_{platform}_{start_time:%Y%m%dT%H%M%SZ}_{end_time:%Y%m%dT%H%M%SZ}_{processing_mode}_{disposition_mode}_{creation_time}_{version_int:04d}.nc'


**Example**:

Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene

    filenames = ['data/AVHRR-GAC_FDR_1C_N06_19810330T042358Z_19810330T060903Z_R_O_20200101T000000Z_0100.nc']
    scn = Scene(reader='avhrr_l1c_eum_gac_fdr_nc', filenames=filenames)
    scn.load(['brightness_temperature_channel_4'])
    scn['brightness_temperature_channel_4']


Output:

.. code-block:: none

    <xarray.DataArray 'brightness_temperature_channel_4' (y: 11, x: 409)>
    dask.array<open_dataset-55ffbf3623b32077c67897f4283640a5brightness_temperature_channel_4, shape=(11, 409),
        dtype=float32, chunksize=(11, 409), chunktype=numpy.ndarray>
    Coordinates:
      * x          (x) int16 0 1 2 3 4 5 6 7 8 ... 401 402 403 404 405 406 407 408
      * y          (y) int64 0 1 2 3 4 5 6 7 8 9 10
        acq_time   (y) datetime64[ns] dask.array<chunksize=(11,), meta=np.ndarray>
        longitude  (y, x) float64 dask.array<chunksize=(11, 409), meta=np.ndarray>
        latitude   (y, x) float64 dask.array<chunksize=(11, 409), meta=np.ndarray>
    Attributes:
        start_time:                            1981-03-30 04:23:58
        end_time:                              1981-03-30 06:09:03
        calibration:                           brightness_temperature
        modifiers:                             ()
        resolution:                            1050
        standard_name:                         toa_brightness_temperature
        units:                                 K
        wavelength:                            10.8 µm (10.3-11.3 µm)
        Conventions:                           CF-1.8 ACDD-1.3
        comment:                               Developed in cooperation with EUME...
        creator_email:                         ops@eumetsat.int
        creator_name:                          EUMETSAT
        creator_url:                           https://www.eumetsat.int/
        date_created:                          2020-09-14T10:50:51.073707
        disposition_mode:                      O
        gac_filename:                          NSS.GHRR.NA.D81089.S0423.E0609.B09...
        geospatial_lat_max:                    89.95386902434623
        geospatial_lat_min:                    -89.97581969005503
        geospatial_lat_resolution:             1050 meters
        geospatial_lat_units:                  degrees_north
        geospatial_lon_max:                    179.99952992568998
        geospatial_lon_min:                    -180.0
        geospatial_lon_resolution:             1050 meters
        geospatial_lon_units:                  degrees_east
        ground_station:                        GC
        id:                                    DOI:10.5676/EUM/AVHRR_GAC_L1C_FDR/...
        institution:                           EUMETSAT
        instrument:                            Earth Remote Sensing Instruments >...
        keywords:                              ATMOSPHERE > ATMOSPHERIC RADIATION...
        keywords_vocabulary:                   GCMD Science Keywords, Version 9.1
        licence:                               EUMETSAT data policy https://www.e...
        naming_authority:                      int.eumetsat
        orbit_number_end:                      9123
        orbit_number_start:                    9122
        orbital_parameters_tle:                ['1 11416U 79057A   81090.16350942...
        platform:                              Earth Observation Satellites > NOA...
        processing_level:                      1C
        processing_mode:                       R
        product_version:                       1.0.0
        references:                            Devasthale, A., M. Raspaud, C. Sch...
        source:                                AVHRR GAC Level 1 Data
        standard_name_vocabulary:              CF Standard Name Table v73
        summary:                               Fundamental Data Record (FDR) of m...
        sun_earth_distance_correction_factor:  0.9975244779999585
        time_coverage_end:                     19820803T003900Z
        time_coverage_start:                   19800101T000000Z
        title:                                 AVHRR GAC L1C FDR
        version_calib_coeffs:                  PATMOS-x, v2017r1
        version_pygac:                         1.4.0
        version_pygac_fdr:                     0.1.dev107+gceb7b26.d20200910
        version_satpy:                         0.21.1.dev894+g5cf76e6
        history:                               Created by pytroll/satpy on 2020-0...
        name:                                  brightness_temperature_channel_4
        _satpy_id:                             DataID(name='brightness_temperatur...
        ancillary_variables:                   []

"""
import itertools
import logging

import xarray as xr
from pyresample import AreaDefinition

import satpy.cf.decoding
from satpy.dataset.dataid import WavelengthRange
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


class SatpyCFFileHandler(BaseFileHandler):
    """File handler for Satpy's CF netCDF files."""

    def __init__(self, filename, filename_info, filetype_info, numeric_name_prefix="CHANNEL_"):
        """Initialize file handler."""
        super().__init__(filename, filename_info, filetype_info)
        self.engine = None
        self._numeric_name_prefix = numeric_name_prefix

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def sensor_names(self):
        """Get sensor set."""
        sensors = set()
        for _, ds_info in self.available_datasets():
            try:
                sensors.add(ds_info["sensor"])
            except KeyError:
                continue
        return sensors

    def available_datasets(self, configured_datasets=None):
        """Add information of available datasets."""
        existing = self._existing_datasets(configured_datasets=configured_datasets)
        dynamic = self._dynamic_datasets()
        coordinates = self._coordinate_datasets()
        for dataset_available, dataset_info in itertools.chain(existing, dynamic, coordinates):
            yield dataset_available, dataset_info

    def _existing_datasets(self, configured_datasets=None):
        """Add information of existing datasets."""
        for is_avail, ds_info in (configured_datasets or []):
            yield is_avail, ds_info

    def fix_modifier_attr(self, ds_info):
        """Fix modifiers attribute."""
        # Empty modifiers are read as [], which causes problems later
        if "modifiers" in ds_info and len(ds_info["modifiers"]) == 0:
            ds_info["modifiers"] = ()
        try:
            try:
                ds_info["modifiers"] = tuple(ds_info["modifiers"].split(" "))
            except AttributeError:
                pass
        except KeyError:
            pass

    def _assign_ds_info(self, var_name, val):
        """Assign ds_info."""
        ds_info = dict(val.attrs)
        ds_info["file_type"] = self.filetype_info["file_type"]
        ds_info["name"] = ds_info["nc_store_name"] = var_name
        if "original_name" in ds_info:
            ds_info["name"] = ds_info["original_name"]
        elif self._numeric_name_prefix and var_name.startswith(self._numeric_name_prefix):
            ds_info["name"] = var_name.replace(self._numeric_name_prefix, "")
        try:
            ds_info["wavelength"] = WavelengthRange.from_cf(ds_info["wavelength"])
        except KeyError:
            pass
        return ds_info

    def _dynamic_datasets(self):
        """Add information of dynamic datasets."""
        nc = xr.open_dataset(self.filename, engine=self.engine)
        # get dynamic variables known to this file (that we created)
        for var_name, val in nc.data_vars.items():
            ds_info = self._assign_ds_info(var_name, val)
            self.fix_modifier_attr(ds_info)
            yield True, ds_info

    def _coordinate_datasets(self, configured_datasets=None):
        """Add information of coordinate datasets."""
        nc = xr.open_dataset(self.filename, engine=self.engine)
        for var_name, val in nc.coords.items():
            ds_info = dict(val.attrs)
            ds_info["file_type"] = self.filetype_info["file_type"]
            ds_info["name"] = var_name
            self.fix_modifier_attr(ds_info)
            yield True, ds_info

    def _compare_attr(self, _ds_id_dict, key, data):
        if key in ["name", "modifiers"]:
            return True
        elif key == "wavelength":
            return _ds_id_dict[key] == WavelengthRange.from_cf(data.attrs[key])
        else:
            return data.attrs[key] == _ds_id_dict[key]

    def _dataid_attrs_equal(self, ds_id, data):
        _ds_id_dict = ds_id.to_dict()
        for key in _ds_id_dict:
            try:
                if not self._compare_attr(_ds_id_dict, key, data):
                    return False
            except KeyError:
                pass
        return True

    def get_dataset(self, ds_id, ds_info):
        """Get dataset."""
        logger.debug("Getting data for: %s", ds_id["name"])
        nc = xr.open_dataset(self.filename, engine=self.engine,
                             chunks={"y": CHUNK_SIZE, "x": CHUNK_SIZE})
        name = ds_info.get("nc_store_name", ds_id["name"])
        data = nc[ds_info.get("file_key", name)]
        if not self._dataid_attrs_equal(ds_id, data):
            return
        if name != ds_id["name"]:
            data = data.rename(ds_id["name"])
        data.attrs.update(nc.attrs)  # For now add global attributes to all datasets
        data.attrs = satpy.cf.decoding.decode_attrs(data.attrs)
        return data

    def get_area_def(self, dataset_id):
        """Get area definition from CF complient netcdf."""
        try:
            area = AreaDefinition.from_cf(self.filename)
            return area
        except ValueError:
            # No CF compliant projection information was found in the netcdf file or
            # file contains 2D lat/lon arrays. To fall back to generating a SwathDefinition
            # with the yaml_reader NotImplementedError is raised.
            logger.debug("No AreaDefinition to load from nc file. Falling back to SwathDefinition.")
            raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Advanced Himawari Imager (AHI) gridded format data reader.

This data comes in a flat binary format on a fixed grid, and needs to have
calibration coefficients applied to it in order to retrieve reflectance or BT.
LUTs can be downloaded at: ftp://hmwr829gr.cr.chiba-u.ac.jp/gridded/FD/support/

This data is gridded from the original Himawari geometry. To our knowledge,
only full disk grids are available, not for the Meso or Japan rapid scans.

References:
 - AHI gridded data website:
        http://www.cr.chiba-u.jp/databases/GEO/H8_9/FD/index_jp.html


"""

import logging
import os

import dask.array as da
import numpy as np
import xarray as xr
from platformdirs import AppDirs
from pyresample import geometry

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.utils import unzip_file
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()

# Hardcoded address of the reflectance and BT look-up tables
AHI_REMOTE_LUTS = "http://www.cr.chiba-u.jp/databases/GEO/H8_9/FD/count2tbb_v102.tgz"

# Full disk image sizes for each spatial resolution
AHI_FULLDISK_SIZES = {0.005: {"x_size": 24000,
                              "y_size": 24000},
                      0.01: {"x_size": 12000,
                             "y_size": 12000},
                      0.02: {"x_size": 6000,
                             "y_size": 6000}}

# Geographic extent of the full disk area in degrees
AHI_FULLDISK_EXTENT = [85., -60., 205., 60.]

# Resolutions of each channel type
AHI_CHANNEL_RES = {"vis": 0.01,
                   "ext": 0.005,
                   "sir": 0.02,
                   "tir": 0.02}

# List of LUT filenames
AHI_LUT_NAMES = ["ext.01", "vis.01", "vis.02", "vis.03",
                 "sir.01", "sir.02", "tir.01", "tir.02",
                 "tir.03", "tir.04", "tir.05", "tir.06",
                 "tir.07", "tir.08", "tir.09", "tir.10"]

logger = logging.getLogger("ahi_grid")


class AHIGriddedFileHandler(BaseFileHandler):
    """AHI gridded format reader.

    This data is flat binary, big endian unsigned short.
    It covers the region 85E -> 205E, 60N -> 60S at variable resolution:
    - 0.005 degrees for Band 3
    - 0.01 degrees for Bands 1, 2 and 4
    - 0.02 degrees for all other bands.
    These are approximately equivalent to 0.5, 1 and 2km.

    Files can either be zipped with bz2 compression (like the HSD format
    data), or can be uncompressed flat binary.
    """

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(AHIGriddedFileHandler, self).__init__(filename, filename_info,
                                                    filetype_info)
        self._unzipped = unzip_file(self.filename)
        # Assume file is not zipped
        if self._unzipped:
            # But if it is, set the filename to point to unzipped temp file
            self.filename = self._unzipped
        # Get the band name, needed for finding area and dimensions
        self.product_name = filetype_info["file_type"]
        self.areaname = filename_info["area"]
        self.sensor = "ahi"
        self.res = AHI_CHANNEL_RES[self.product_name[:3]]
        if self.areaname == "fld":
            self.nlines = AHI_FULLDISK_SIZES[self.res]["y_size"]
            self.ncols = AHI_FULLDISK_SIZES[self.res]["x_size"]
        else:
            raise NotImplementedError("Only full disk data is supported.")

        # Set up directory path for the LUTs
        app_dirs = AppDirs("ahi_gridded_luts", "satpy", "1.0.2")
        self.lut_dir = os.path.expanduser(app_dirs.user_data_dir) + "/"
        self.area = None

    def __del__(self):
        """Delete the object."""
        if self._unzipped and os.path.exists(self.filename):
            os.remove(self.filename)

    def _load_lut(self):
        """Determine if LUT is available and, if not, download it."""
        # First, check that the LUT is available. If not, download it.
        lut_file = self.lut_dir + self.product_name
        if not os.path.exists(lut_file):
            self._get_luts()
        try:
            # Load file, it has 2 columns: DN + Refl/BT. We only need latter.
            lut = np.loadtxt(lut_file)[:, 1]
        except FileNotFoundError:
            raise FileNotFoundError("No LUT file found:", lut_file)
        return lut

    def _calibrate(self, data):
        """Load calibration from LUT and apply."""
        lut = self._load_lut()

        # LUT may truncate NaN values, so manually set those in data
        lut_len = len(lut)
        data = np.where(data < lut_len - 1, data, np.nan)
        return lut[data.astype(np.uint16)]

    @staticmethod
    def _download_luts(file_name):
        """Download LUTs from remote server."""
        import shutil
        import urllib

        # Set up an connection and download
        with urllib.request.urlopen(AHI_REMOTE_LUTS) as response:  # nosec
            with open(file_name, "wb") as out_file:
                shutil.copyfileobj(response, out_file)

    @staticmethod
    def _untar_luts(tarred_file, outdir):
        """Uncompress downloaded LUTs, which are a tarball."""
        import tarfile
        tar = tarfile.open(tarred_file)
        tar.extractall(outdir, filter="data")
        tar.close()
        os.remove(tarred_file)

    def _get_luts(self):
        """Download the LUTs needed for count->Refl/BT conversion."""
        import pathlib
        import shutil

        from satpy import config

        # Check that the LUT directory exists
        pathlib.Path(self.lut_dir).mkdir(parents=True, exist_ok=True)

        logger.info("Download AHI LUTs files and store in directory %s",
                    self.lut_dir)
        tempdir = config["tmp_dir"]
        fname = os.path.join(tempdir, "tmp.tgz")
        # Download the LUTs
        self._download_luts(fname)

        # The file is tarred, untar and remove the downloaded file
        self._untar_luts(fname, tempdir)

        lut_dl_dir = os.path.join(tempdir, "count2tbb_v102/")

        # Loop over the LUTs and copy to the correct location
        for lutfile in AHI_LUT_NAMES:
            shutil.move(os.path.join(lut_dl_dir, lutfile), os.path.join(self.lut_dir, lutfile))
        shutil.rmtree(lut_dl_dir)

    def get_dataset(self, key, info):
        """Get the dataset."""
        return self.read_band(key, info)

    def get_area_def(self, dsid):
        """Get the area definition.

        This is fixed, but not defined in the file. So we must
        generate it ourselves with some assumptions.
        """
        if self.areaname == "fld":
            area_extent = AHI_FULLDISK_EXTENT
        else:
            raise NotImplementedError("Reader only supports full disk data.")

        proj_param = "EPSG:4326"

        area = geometry.AreaDefinition("gridded_himawari",
                                       "A gridded Himawari area",
                                       "longlat",
                                       proj_param,
                                       self.ncols,
                                       self.nlines,
                                       area_extent)
        self.area = area

        return area

    def _read_data(self, fp_):
        """Read raw binary data from file."""
        return da.from_array(np.memmap(self.filename,
                                       offset=fp_.tell(),
                                       dtype=">u2",
                                       shape=(self.nlines, self.ncols),
                                       mode="r"),
                             chunks=CHUNK_SIZE)

    def read_band(self, key, info):
        """Read the data."""
        with open(self.filename, "rb") as fp_:
            res = self._read_data(fp_)

        # Calibrate
        res = self.calibrate(res, key["calibration"])

        # Update metadata
        new_info = dict(
            units=info["units"],
            standard_name=info["standard_name"],
            wavelength=info["wavelength"],
            resolution=info["resolution"],
            id=key,
            name=key["name"],
            sensor=self.sensor,
        )
        res = xr.DataArray(res, attrs=new_info, dims=["y", "x"])
        return res

    def calibrate(self, data, calib):
        """Calibrate the data."""
        if calib == "counts":
            return data
        if calib == "reflectance" or calib == "brightness_temperature":
            return self._calibrate(data)
        raise NotImplementedError("ERROR: Unsupported calibration.",
                                  "Only counts, reflectance and ",
                                  "brightness_temperature calibration",
                                  "are supported.")



# Copyright (c) 2019-2023 Satpy developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Advance Baseline Imager NOAA Level 2+ products reader.

The files read by this reader are described in the official PUG document:
    https://www.goes-r.gov/products/docs/PUG-L2+-vol5.pdf

Data Quality Filtering
^^^^^^^^^^^^^^^^^^^^^^

Some variables can be filtered based on Data Quality Flags (DQF) in the
data files. The flag meanings to retain, where others are marked as invalid,
are specified with the file handler keyword argument ``filters=["good_quality_qf"]``.
These invalid values are marked by NaN in the returned data for floating point arrays.
The values in this list
must match the entries in the "flag_meanings" attribute of the "DQF"
variable in the file. By default no filtering is applied.

"""

import logging
import operator
from functools import reduce

import numpy as np
import xarray as xr

from satpy.readers.core.abi import NC_ABI_BASE

LOG = logging.getLogger(__name__)


class NC_ABI_L2(NC_ABI_BASE):
    """Reader class for NOAA ABI l2+ products in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info,
                 filters: list[str] | None = None, **kwargs):
        """Initialize file handler and store filter_sst state."""
        super().__init__(filename, filename_info, filetype_info, **kwargs)
        self.filters = filters or []

    def get_dataset(self, key, info):
        """Load a dataset."""
        var = info["file_key"]
        if self.filetype_info["file_type"] == "abi_l2_mcmip":
            var += "_" + key["name"]
        LOG.debug("Reading in get_dataset %s.", var)
        variable = self[var]
        variable.attrs.update(key.to_dict())
        self._update_data_arr_with_filename_attrs(variable)
        self._remove_problem_attrs(variable)
        variable = self._filter_dqf(variable)

        # convert to satpy standard units
        if variable.attrs["units"] == "1" and key.get("calibration") == "reflectance":
            variable *= 100.0
            variable.attrs["units"] = "%"

        return variable

    def _update_data_arr_with_filename_attrs(self, variable):
        _units = variable.attrs["units"] if "units" in variable.attrs else None
        variable.attrs.update({
            "platform_name": self.platform_name,
            "sensor": self.sensor,
            "units": _units,
            "orbital_parameters": {
                "satellite_nominal_latitude": float(self.nc["nominal_satellite_subpoint_lat"]),
                "satellite_nominal_longitude": float(self.nc["nominal_satellite_subpoint_lon"]),
                "satellite_nominal_altitude": float(self.nc["nominal_satellite_height"]) * 1000.,
            },
        })
        self._convert_flag_attrs(variable)

        # add in information from the filename that may be useful to the user
        for attr in ("scene_abbr", "scan_mode", "platform_shortname"):
            variable.attrs[attr] = self.filename_info.get(attr)

        # add in information hardcoded in the filetype YAML
        for attr in ("observation_type",):
            if attr in self.filetype_info:
                variable.attrs[attr] = self.filetype_info[attr]

        # copy global attributes to metadata
        for attr in ("scene_id", "orbital_slot", "instrument_ID", "production_site", "timeline_ID"):
            variable.attrs[attr] = self.nc.attrs.get(attr)

    @staticmethod
    def _convert_flag_attrs(variable: xr.DataArray) -> None:
        if "flag_meanings" in variable.attrs:
            variable.attrs["flag_meanings"] = variable.attrs["flag_meanings"].split(" ")
        if "flag_values" in variable.attrs:
            variable.attrs["flag_values"] = [int(val) for val in variable.attrs["flag_values"]]

    @staticmethod
    def _remove_problem_attrs(variable):
        # remove attributes that could be confusing later
        if not np.issubdtype(variable.dtype, np.integer):
            # integer fields keep the _FillValue
            variable.attrs.pop("_FillValue", None)
        variable.attrs.pop("scale_factor", None)
        variable.attrs.pop("add_offset", None)
        variable.attrs.pop("valid_range", None)
        variable.attrs.pop("_Unsigned", None)
        variable.attrs.pop("valid_range", None)
        variable.attrs.pop("ancillary_variables", None)  # Can't currently load DQF

    def _filter_dqf(self, variable: xr.DataArray) -> xr.DataArray:
        if "DQF" not in self:
            return variable
        dqf = self["DQF"]
        if "flag_meanings" not in dqf.attrs or "flag_values" not in dqf.attrs:
            return variable

        self._convert_flag_attrs(dqf)
        flag_dict = dict(zip(dqf.attrs["flag_meanings"], dqf.attrs["flag_values"], strict=True))
        values_to_keep = [flag_dict[flag_name] for flag_name in self.filters if flag_name in flag_dict]
        if not values_to_keep:
            # user didn't specify any filters for this variable
            return variable

        LOG.debug(f"Filtering {variable.attrs['name']} for DQF flags {values_to_keep!r}")
        good_mask = reduce(operator.or_, [dqf == flag_value for flag_value in values_to_keep])
        return variable.where(good_mask)

    def available_datasets(self, configured_datasets=None):
        """Add resolution to configured datasets."""
        for is_avail, ds_info in (configured_datasets or []):
            # some other file handler knows how to load this
            # don't override what they've done
            if is_avail is not None:
                yield is_avail, ds_info
            matches = self.file_type_matches(ds_info["file_type"])
            if matches:
                # we have this dataset
                resolution = self.spatial_resolution_to_number()
                new_info = ds_info.copy()
                new_info.setdefault("resolution", resolution)
                yield True, ds_info
            elif is_avail is None:
                # we don't know what to do with this
                # see if another future file handler does
                yield is_avail, ds_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to VIRR (Visible and Infra-Red Radiometer) level 1b format.

The file format is HDF5. Important attributes:

    - Latitude
    - Longitude
    - SolarZenith
    - EV_Emissive
    - EV_RefSB
    - Emissive_Radiance_Offsets
    - Emissive_Radiance_Scales
    - RefSB_Cal_Coefficients
    - RefSB_Effective_Wavelength
    - Emmisive_Centroid_Wave_Number

Supported satellites:

    - FY-3B and FY-3C.

For more information:

    - https://www.wmo-sat.info/oscar/instruments/view/607.

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
from pyspectral.blackbody import blackbody_wn_rad2temp as rad2temp

from satpy.readers.core.hdf5 import HDF5FileHandler

LOG = logging.getLogger(__name__)

# PROVIDED BY NIGEL ATKINSON - 2013
# FY3B_REF_COEFFS = [
#     0.12640, -1.43200,  #channel1#
#     0.13530, -1.62360,  #channel2#
#     0.09193, -2.48207,  #channel6#
#     0.07480, -0.90980,  #channel7#
#     0.07590, -0.91080,  #channel8#
#     0.07460, -0.89520,  #channel9#
#     0.06300, -0.76280]  #channel10#
# CMA - 2015 - http://www.nsmc.org.cn/en/NSMC/Contents/100089.html
FY3B_REF_COEFFS = [
    0.1264, -1.4320,
    0.1353, -1.6236,
    0.0919, -2.4821,
    0.0938, -1.1494,
    0.0857, -1.0280,
    0.0803, -0.9636,
    0.0630, -0.7628]


class VIRR_L1B(HDF5FileHandler):
    """VIRR Level 1b reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Open file and perform initial setup."""
        super(VIRR_L1B, self).__init__(filename, filename_info, filetype_info)
        LOG.debug("day/night flag for {0}: {1}".format(filename, self["/attr/Day Or Night Flag"]))
        self.geolocation_prefix = filetype_info["geolocation_prefix"]
        self.platform_id = filename_info["platform_id"]
        self.l1b_prefix = "Data/"
        self.wave_number = "Emissive_Centroid_Wave_Number"
        # Else filename_info['platform_id'] == FY3C.
        if filename_info["platform_id"] == "FY3B":
            self.l1b_prefix = ""
            self.wave_number = "Emmisive_Centroid_Wave_Number"

    def get_dataset(self, dataset_id, ds_info):
        """Create DataArray from file content for `dataset_id`."""
        file_key = self.geolocation_prefix + ds_info.get("file_key", dataset_id["name"])
        if self.platform_id == "FY3B":
            file_key = file_key.replace("Data/", "")
        data = self[file_key]
        band_index = ds_info.get("band_index")
        valid_range = data.attrs.pop("valid_range", None)
        if isinstance(valid_range, np.ndarray):
            valid_range = valid_range.tolist()
        if band_index is not None:
            data = data[band_index]
            if valid_range:
                data = data.where((data >= valid_range[0]) &
                                  (data <= valid_range[1]))
            if "Emissive" in file_key:
                self._calibrate_emissive(data, band_index)
            elif "RefSB" in file_key:
                data = self._calibrate_reflective(data, band_index)
        else:
            slope = self._correct_slope(self[file_key + "/attr/Slope"])
            intercept = self[file_key + "/attr/Intercept"]

            if valid_range:
                data = data.where((data >= valid_range[0]) &
                                  (data <= valid_range[1]))
            data = data * slope + intercept
        new_dims = {old: new for old, new in zip(data.dims, ("y", "x"))}
        data = data.rename(new_dims)
        # use lowercase sensor name to be consistent with the rest of satpy
        data.attrs.update({"platform_name": self["/attr/Satellite Name"],
                           "sensor": self["/attr/Sensor Identification Code"].lower()})
        data.attrs.update(ds_info)
        units = self.get(file_key + "/attr/units")
        if units is not None and str(units).lower() != "none":
            data.attrs.update({"units": self.get(file_key + "/attr/units")})
        elif data.attrs.get("calibration") == "reflectance":
            data.attrs.update({"units": "%"})
        else:
            data.attrs.update({"units": "1"})
        return data

    def _calibrate_reflective(self, data, band_index):
        if self.platform_id == "FY3B":
            coeffs = da.from_array(FY3B_REF_COEFFS, chunks=-1)
        else:
            coeffs = self["/attr/RefSB_Cal_Coefficients"]
        slope = self._correct_slope(coeffs[0::2])
        intercept = coeffs[1::2]
        data = data * slope[band_index] + intercept[band_index]
        return data

    def _calibrate_emissive(self, data, band_index):
        slope = self._correct_slope(self[self.l1b_prefix + "Emissive_Radiance_Scales"].
                                    data[:, band_index][:, np.newaxis])
        intercept = self[self.l1b_prefix + "Emissive_Radiance_Offsets"].data[:, band_index][:, np.newaxis]
        # Converts cm^-1 (wavenumbers) and (mW/m^2)/(str/cm^-1) (radiance data)
        # to SI units m^-1, mW*m^-3*str^-1.
        wave_number = self["/attr/" + self.wave_number][band_index] * 100
        bt_data = rad2temp(wave_number, (data.data * slope + intercept) * 1e-5)
        if isinstance(bt_data, np.ndarray):
            # old versions of pyspectral produce numpy arrays
            data.data = da.from_array(bt_data, chunks=data.data.chunks)
        else:
            # new versions of pyspectral can do dask arrays
            data.data = bt_data

    def _correct_slope(self, slope):
        # 0 slope is invalid. Note: slope can be a scalar or array.
        return da.where(slope == 0, 1, slope)

    @property
    def start_time(self):
        """Get starting observation time."""
        start_time = self["/attr/Observing Beginning Date"] + "T" + self["/attr/Observing Beginning Time"] + "Z"
        return dt.datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """Get ending observation time."""
        end_time = self["/attr/Observing Ending Date"] + "T" + self["/attr/Observing Ending Time"] + "Z"
        return dt.datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%S.%fZ")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from urllib.error import URLError
from urllib.request import urlopen

try:
    import gcsfs
except ImportError:
    gcsfs = None

LOG = logging.getLogger(__name__)


def is_google_cloud_instance():
    """Check if we are on a GCP virtual machine."""
    try:
        return urlopen("http://metadata.google.internal").headers.get("Metadata-Flavor") == "Google"  # nosec
    except URLError:
        return False


def get_bucket_files(glob_pattern, base_dir, force=False, pattern_slice=None):
    """Download files from Google Cloud Storage.

    Args:
        glob_pattern (str or list): Glob pattern string or series of patterns
            used to search for on Google Cloud Storage. The pattern should
            include the "gs://" protocol prefix. If a list of lists, then the
            results of each sublist pattern are concatenated and the result is
            treated as one pattern result. This is important for things like
            ``pattern_slice`` and complicated glob patterns not supported by
            GCP.
        base_dir (str): Root directory to place downloaded files on the local
            system.
        force (bool): Force re-download of data regardless of its existence on
            the local system. Warning: May delete non-demo files stored in
            download directory.
        pattern_slice (slice): Slice object to limit the number of files
            returned by each glob pattern.

    """
    if pattern_slice is None:
        pattern_slice = slice(None)
    if gcsfs is None:
        raise RuntimeError("Missing 'gcsfs' dependency for GCS download.")
    if not os.path.isdir(base_dir):
        # it is the caller's responsibility to make this
        raise OSError("Directory does not exist: {}".format(base_dir))

    if isinstance(glob_pattern, str):
        glob_pattern = [glob_pattern]

    fs = gcsfs.GCSFileSystem(token="anon")  # nosec
    filenames = []
    for gp in glob_pattern:
        # handle multiple glob patterns being treated as one pattern
        # for complicated patterns that GCP can't handle
        if isinstance(gp, str):
            glob_results = list(fs.glob(gp))
        else:
            # flat list of results
            glob_results = [fn for pat in gp for fn in fs.glob(pat)]
        filenames.extend(_download_gcs_files(glob_results[pattern_slice], fs, base_dir, force))

    if not filenames:
        raise OSError("No files could be found or downloaded.")
    return filenames


def _download_gcs_files(globbed_files, fs, base_dir, force):
    filenames = []
    for fn in globbed_files:
        ondisk_fn = os.path.basename(fn)
        ondisk_pathname = os.path.join(base_dir, ondisk_fn)
        filenames.append(ondisk_pathname)

        if force and os.path.isfile(ondisk_pathname):
            os.remove(ondisk_pathname)
        elif os.path.isfile(ondisk_pathname):
            LOG.info("Found existing: {}".format(ondisk_pathname))
            continue
        LOG.info("Downloading: {}".format(ondisk_pathname))
        fs.get("gs://" + fn, ondisk_pathname)
    return filenames



"""Tests for the CAMEL L3 netCDF reader."""

import datetime as dt
import os

import numpy as np
import pytest
import xarray as xr

from satpy.readers.camel_l3_nc import CAMELL3NCFileHandler
from satpy.tests.utils import make_dataid

rng = np.random.default_rng()
ndvi_data = rng.integers(0, 1000, (3600, 7200), dtype=np.int16)
emis_data = rng.integers(0, 1000, (3600, 7200, 5), dtype=np.int16)

lon_data = np.arange(-180, 180, 0.05)
lat_data = np.arange(-90, 90, 0.05)

start_time = dt.datetime(2023, 8, 1, 0, 0, 0)
end_time = dt.datetime(2023, 9, 1, 0, 0, 0)

fill_val = -999
scale_val = 0.001

dimensions = {"longitude": 7200, "latitude": 3600, "spectra": 13}

exp_ext = (-180.0, -90.0, 180.0, 90.0)

global_attrs = {"time_coverage_start": start_time.strftime("%Y-%m-%d %H:%M:%SZ"),
                "time_coverage_end": end_time.strftime("%Y-%m-%d %H:%M:%SZ"),
                "geospatial_lon_resolution": "0.05 degree grid ",
                "geospatial_lat_resolution": "0.05 degree grid ",
                }

bad_attrs1 = global_attrs.copy()
bad_attrs1["geospatial_lon_resolution"] = "0.1 degree grid "
bad_attrs2 = global_attrs.copy()
bad_attrs2["geospatial_lat_resolution"] = "0.1 degree grid "


def _make_ds(the_attrs, tmp_factory):
    """Make a dataset for use in tests."""
    fname = f'{tmp_factory.mktemp("data")}/CAM5K30EM_emis_202308_V003.nc'
    ds = xr.Dataset({"aster_ndvi": (["Rows", "Columns"], ndvi_data),
                     "camel_emis": (["latitude", "longitude", "spectra"], emis_data)},
                    coords={"latitude": (["Rows"], lat_data),
                            "longitude": (["Columns"], lon_data)},
                    attrs=the_attrs)
    ds.to_netcdf(fname)
    return fname


def camel_l3_filehandler(fname):
    """Instantiate a Filehandler."""
    fileinfo = {"start_period": "202308",
                "version": "003"}
    filetype = None
    fh = CAMELL3NCFileHandler(fname, fileinfo, filetype)
    return fh


@pytest.fixture(scope="module")
def camel_filename(tmp_path_factory):
    """Create a fake camel l3 file."""
    fname = _make_ds(global_attrs, tmp_path_factory)
    yield fname
    os.unlink(fname)


@pytest.fixture(scope="module")
def camel_filename_bad1(tmp_path_factory):
    """Create a fake camel l3 file."""
    fname = _make_ds(bad_attrs1, tmp_path_factory)
    yield fname
    os.unlink(fname)


@pytest.fixture(scope="module")
def camel_filename_bad2(tmp_path_factory):
    """Create a fake camel l3 file."""
    fname = _make_ds(bad_attrs2, tmp_path_factory)
    yield fname
    os.unlink(fname)


def test_startend(camel_filename):
    """Test start and end times are set correctly."""
    fh = camel_l3_filehandler(camel_filename)
    assert fh.start_time == start_time
    assert fh.end_time == end_time


def test_camel_l3_area_def(camel_filename, caplog):
    """Test reader handles area definition correctly."""
    from pyproj.crs import CRS

    # Check case where input data is correct size.
    fh = camel_l3_filehandler(camel_filename)
    ndvi_id = make_dataid(name="aster_ndvi")
    area_def = fh.get_area_def(ndvi_id)
    assert area_def.width == dimensions["longitude"]
    assert area_def.height == dimensions["latitude"]
    assert np.allclose(area_def.area_extent, exp_ext)

    assert area_def.crs == CRS.from_epsg(4326)


def test_bad_longitude(camel_filename_bad1):
    """Check case where longitude grid is not correct."""
    with pytest.raises(ValueError, match="Only 0.05 degree grid data is supported."):
        camel_l3_filehandler(camel_filename_bad1)


def test_bad_latitude(camel_filename_bad2):
    """Check case where latitude grid is not correct."""
    with pytest.raises(ValueError, match="Only 0.05 degree grid data is supported."):
        camel_l3_filehandler(camel_filename_bad2)


def test_load_ndvi_data(camel_filename):
    """Test that data is loaded successfully."""
    fh = camel_l3_filehandler(camel_filename)
    ndvi_id = make_dataid(name="aster_ndvi")
    ndvi = fh.get_dataset(ndvi_id, {"file_key": "aster_ndvi"})
    assert np.allclose(ndvi.data, ndvi_data)


def test_load_emis_data(camel_filename):
    """Test that data is loaded successfully."""
    fh = camel_l3_filehandler(camel_filename)
    emis_id = make_dataid(name="camel_emis")

    # This is correct data
    emis = fh.get_dataset(emis_id, {"file_key": "camel_emis", "band_id": 2})
    assert np.allclose(emis.data, emis_data[:, :, 2])

    # This will fail as we are requesting a band too high data
    with pytest.raises(ValueError, match="Band id requested is larger than dataset."):
        fh.get_dataset(emis_id, {"file_key": "camel_emis", "band_id": 12})
